// -*- Mode: Go; indent-tabs-mode: t -*-
//
// Copyright (C) 2018 Canonical Ltd
// Copyright (C) 2018 IOTech Ltd
//
// SPDX-License-Identifier: Apache-2.0

// This package provides a simple example implementation of
// a ProtocolDriver interface.
//
package driver

import (
	"fmt"
	"time"
	"encoding/json"

	ds_models "github.com/edgexfoundry/device-wise-agent/pkg/models"
	"github.com/edgexfoundry/edgex-go/pkg/clients/logging"
	"github.com/edgexfoundry/edgex-go/pkg/models"
	MQTT "github.com/eclipse/paho.mqtt.golang"
)


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Wise-Agent handshake protocol data format 
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* 2.0 */
const (
	INFOACK_TOPIC = "/wisepaas/device/+/agentinfoack"
	DEF_AGENTACT_TOPIC = "/wisepaas/%s/+/agentactionack"
	DEF_AGENTREPORT_TOPIC = "/wisepaas/device/+/devinfoack"
	DEF_CALLBACKREQ_TOPIC = "/wisepaas/RMM/%s/agentactionreq"
)


const (
	INFOACK_TOPIC_3X = "/cagent/admin/+/agentinfoack"
	DEF_AGENTACT_TOPIC_3X = "/cagent/admin/+/agentactionreq"
	DEF_AGENTREPORT_TOPIC_3X = "/cagent/admin/+/deviceinfo"
	DEF_CALLBACKREQ_TOPIC_3X = "/cagent/admin/%s/agentcallbackreq"
)


const (
	WISE_UNKNOWN_CMD = 0
	WISE_AGENTINFO_CMD = 21
	WISE_UPDATE_CAGENT_REQ = 111
	WISE_UPDATE_CAGENT_REP
	WISE_CAGENT_RENAME_REQ = 113
	WISE_CAGENT_RENAME_REP
	WISE_CAGENT_OSINFO_REP = 116
	WISE_SERVER_CONTROL_REQ = 125
	WISE_SERVER_CONTROL_REP

	WISE_HEARTBEATRATE_QUERY_REQ = 127
	WISE_HEARTBEATRATE_QUERY_REP = 128
	WISE_HEARTBEATRATE_UPDATE_REQ = 129
	WISE_HEARTBEATRATE_UPDATE_REP = 130

	WISE_GET_SENSOR_DATA_REQ = 523
	WISE_GET_SENSOR_DATA_REP = 524
	WISE_SET_SENSOR_DATA_REQ = 525
	WISE_SET_SENSOR_DATA_REP = 526

	WISE_ERROR_REP = 600

	WISE_GET_CAPABILITY_REQ = 2051
	WISE_GET_CAPABILITY_REP = 2052
	WISE_START_REPORT_REQ = 2053
	WISE_START_REPORT_REP = 2054
	WISE_REPORT_DATA_REP = 2055
	WISE_STOP_REPORT_REQ = 2056
	WISE_STOP_REPORT_REP = 2057
)

const (
	MODULE_PROCESS = "ProcessMonitor"
	MODULE_DATA_SYNC = "DataSync"
)

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Command struct 2.0
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
type AgentHeader struct {
	AgentID     string `json:"agentID"`
	HandlerName string `json:"handlerName"`
	CommCmd     int    `json:"commCmd"`
	Content     interface{} `json:"content,omitempty"`
	SendTS struct {
		Date int64 `json:"$date"`
	} `json:"sendTS"`
}

type ContentAgentInfo struct {
	ParentID    string `json:"parentID"`
	Hostname    string `json:"hostname"`
	Sn          string `json:"sn"`
	Mac         string `json:"mac"`
	Version     string `json:"version"`
	Type        string `json:"type"`
	Tag         string `json:"tag"`
	Product     string `json:"product"`
	Manufacture string `json:"manufacture"`
	Account     string `json:"account"`
	Password    string `json:"password"`
	Status      int    `json:"status"`
}

type ContentStartReport struct {
	AutoUploadIntervalSec int `json:"autoUploadIntervalSec"`
	RequestItems struct {
		ALL struct {
		} `json:"ALL"`
	} `json:"requestItems"`
}

type ContentGetSensor struct {
	SessionID    string `json:"sessionID"`
	SensorIDList struct {
		E []struct {
			N string `json:"n"`
		} `json:"e"`
	} `json:"sensorIDList"`
}

type ContentDataSync struct {
	RecoverProgress struct {
		Bn string `json:"bn"`
		E  []struct {
			N   string `json:"n"`
			V   int    `json:"v"`
			Asm string `json:"asm,omitempty"`
		} `json:"e"`
	} `json:"RecoverProgress"`
	OpTS struct {
		Date int64 `json:"$date"`
	} `json:"opTS"`
	DataFlow string `json:"dataFlow"`
	Seq      string `json:"seq"`
	SrcTs    int64  `json:"srcTs"`
}

type ContentProcess struct {
	SystemMonitorInfo struct {
		E []struct {
			N string `json:"n"`
			V uint64 `json:"v"`
		} `json:"e"`
		Bn      string `json:"bn"`
		Basm    string `json:"basm"`
		Btype   string `json:"btype"`
		Bst     string `json:"bst"`
		Bextend string `json:"bextend"`
	} `json:"System Monitor Info"`
	ProcessMonitorInfo []struct {
		E []struct {
			N  string `json:"n"`
			Sv string `json:"sv,omitempty"`
			V  int    `json:"v,omitempty"`
		} `json:"e"`
		Bn      string `json:"bn"`
		Basm    string `json:"basm"`
		Btype   string `json:"btype"`
		Bst     string `json:"bst"`
		Bextend string `json:"bextend"`
	} `json:"Process Monitor Info"`
	OpTS struct {
		Date int64 `json:"$date"`
	} `json:"opTS"`
	DataFlow string `json:"dataFlow"`
	Seq      string `json:"seq"`
	SrcTs    int64  `json:"srcTs"`
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Command struct 3.3
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
type AgentInfoAck_3X struct {
	SusiCommData struct {
		DevID       string `json:"devID"`
		ParentID    string `json:"parentID"`
		Hostname    string `json:"hostname"`
		Sn          string `json:"sn"`
		Mac         string `json:"mac"`
		Version     string `json:"version"`
		Type        string `json:"type"`
		Product     string `json:"product"`
		Manufacture string `json:"manufacture"`
		Account     string `json:"account"`
		Passwd      string `json:"passwd"`
		Status      int    `json:"status"`
		CommCmd     int    `json:"commCmd"`
		RequestID   int    `json:"requestID"`
		AgentID     string `json:"agentID"`
		HandlerName string `json:"handlerName"`
		SendTS      struct {
			Date int64 `json:"$date"`
		} `json:"sendTS"`
	} `json:"susiCommData"`
}

type RequestCapability_3X struct {
	SusiCommData struct {
		RequestID   int    `json:"requestID"`
		CatalogID   int    `json:"catalogID"`
		CommCmd     int    `json:"commCmd"`
		HandlerName string `json:"handlerName"`
	} `json:"susiCommData"`
}

type RequestStartReport_3X struct {
	SusiCommData struct {
		CatalogID             int `json:"catalogID"`
		AutoUploadIntervalSec int `json:"autoUploadIntervalSec"`
		RequestItems          struct {
			All struct {
			} `json:"All"`
		} `json:"requestItems"`
		HandlerName string `json:"handlerName"`
		CommCmd     int    `json:"commCmd"`
	} `json:"susiCommData"`
}

type CapabilityProcess_3X struct {
	SusiCommData struct {
		InfoSpec struct {
			ProcessMonitor struct {
				SystemMonitorInfo struct {
					E []struct {
						N string `json:"n"`
						V uint64 `json:"v"`
					} `json:"e"`
					Bn      string `json:"bn"`
					Basm    string `json:"basm"`
					Btype   string `json:"btype"`
					Bst     string `json:"bst"`
					Bextend string `json:"bextend"`
				} `json:"System Monitor Info"`
				ProcessMonitorInfo []struct {
					E []struct {
						N  string `json:"n"`
						Sv string `json:"sv,omitempty"`
						V  int    `json:"v,omitempty"`
					} `json:"e"`
					Bn      string `json:"bn"`
					Basm    string `json:"basm"`
					Btype   string `json:"btype"`
					Bst     string `json:"bst"`
					Bextend string `json:"bextend"`
				} `json:"Process Monitor Info"`
			} `json:"ProcessMonitor"`
			OpTS struct {
				Date int64 `json:"$date"`
			} `json:"opTS"`
			DataFlow string `json:"dataFlow"`
			Seq      string `json:"seq"`
			SrcTs    int64  `json:"srcTs"`
		} `json:"infoSpec"`
		CommCmd     int    `json:"commCmd"`
		AgentID     string `json:"agentID"`
		HandlerName string `json:"handlerName"`
		SendTS      struct {
			Date int64 `json:"$date"`
		} `json:"sendTS"`
	} `json:"susiCommData"`
}


type ReportProcess_3X struct {
	SusiCommData struct {
		Data struct {
			ProcessMonitor struct {
				SystemMonitorInfo struct {
					E []struct {
						N string `json:"n"`
						V uint64 `json:"v"`
					} `json:"e"`
					Bn      string `json:"bn"`
					Basm    string `json:"basm"`
					Btype   string `json:"btype"`
					Bst     string `json:"bst"`
					Bextend string `json:"bextend"`
				} `json:"System Monitor Info"`
				ProcessMonitorInfo []struct {
					E []struct {
						N  string `json:"n"`
						Sv string `json:"sv,omitempty"`
						V  int    `json:"v,omitempty"`
					} `json:"e"`
					Bn      string `json:"bn"`
					Basm    string `json:"basm"`
					Btype   string `json:"btype"`
					Bst     string `json:"bst"`
					Bextend string `json:"bextend"`
				} `json:"Process Monitor Info"`
			} `json:"ProcessMonitor"`
			OpTS struct {
				Date int64 `json:"$date"`
			} `json:"opTS"`
			DataFlow string `json:"dataFlow"`
			Seq      string `json:"seq"`
			SrcTs    int64  `json:"srcTs"`
		} `json:"data"`
		CommCmd     int    `json:"commCmd"`
		AgentID     string `json:"agentID"`
		HandlerName string `json:"handlerName"`
		SendTS      struct {
			Date int64 `json:"$date"`
		} `json:"sendTS"`
	} `json:"susiCommData"`
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Driver struct
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
type EdgeXDriver struct {
	lc           logger.LoggingClient
	asyncCh      chan<- *ds_models.AsyncValues
	switchButton bool
	agentDriver  *WiseAgentDriver
}

type AgentDriverData struct {
	cpu uint8
	totalPhysMemKB uint64
	availPhysMemKB uint64
}

type WiseAgentDriver struct {
	mqttClient   MQTT.Client
	data         AgentDriverData
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Wise-Agent handshake protocol implement
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
func PrintMQTT(msg MQTT.Message, data interface{}) {
	fmt.Printf("Get Topic: %s\n", msg.Topic())
	fmt.Printf("Message: %+v\n", data)
}

func (s *EdgeXDriver) SendAgentActionReq(agentID string, commCmd int, content interface{}) (error) {
	request := AgentHeader{
					AgentID:agentID,
					HandlerName:"general", 
					CommCmd:commCmd, 
					Content:content}
	request.SendTS.Date = int64(time.Now().UnixNano() / 1000000)
	payload, _ := json.Marshal(&request)
	topic := fmt.Sprintf(DEF_CALLBACKREQ_TOPIC, agentID)
	token := s.agentDriver.mqttClient.Publish(topic, 0, false, string(payload))
	token.Wait()

	return nil;
}

func (s *WiseAgentDriver) HandleProcess(capRaw *json.RawMessage) {
	capability := ContentProcess{}
	if err := json.Unmarshal([]byte(*capRaw), &capability); err != nil {
		fmt.Println("HandleProcess:", err)
		return
	}

	for _, e := range capability.SystemMonitorInfo.E {
		switch e.N {
		case "CPU Usage":
			s.data.cpu = uint8(e.V)
		case "totalPhysMemKB":
			s.data.totalPhysMemKB = e.V
		case "availPhysMemKB":
			s.data.availPhysMemKB = e.V
		default:
		}
	}
	fmt.Printf("HandleProcess: [%s] CPU=%d, totalPhysMemKB=%d, availPhysMemKB=%d\n",
			MODULE_PROCESS,
			s.data.cpu,
			s.data.totalPhysMemKB,
			s.data.availPhysMemKB)
}

func (s *EdgeXDriver) AgentInfoHandler(client MQTT.Client, msg MQTT.Message) {
	var contentRaw json.RawMessage

	header := AgentHeader { Content: &contentRaw, }
	if err := json.Unmarshal([]byte(msg.Payload()), &header); err != nil {
		fmt.Println("AgentInfoHandler header:", err)
	}

	content := ContentAgentInfo{}
	if err := json.Unmarshal([]byte(contentRaw), &content); err != nil {
		fmt.Println("AgentInfoHandler content:", err)
		return
	}
	
	//PrintMQTT(msg, agentinfoack)
	fmt.Println("AgentInfoHandler: Device ", header.AgentID, " online, MAC=", content.Mac)
	
	// send capabiity requests
	fmt.Println("AgentInfoHandler send capability request")
	s.SendAgentActionReq(header.AgentID, WISE_GET_CAPABILITY_REQ, nil)

	// send start reort
	fmt.Println("AgentInfoHandler send start report")
	reportRequest := ContentStartReport{AutoUploadIntervalSec:10}
	s.SendAgentActionReq(header.AgentID, WISE_START_REPORT_REQ, &reportRequest)
}

func (s *EdgeXDriver) DevInfoHandler(client MQTT.Client, msg MQTT.Message) {
	var contentRaw json.RawMessage
	var content map[string]*json.RawMessage

	header := AgentHeader { Content: &contentRaw, }
	
	// parse header
	if err := json.Unmarshal([]byte(msg.Payload()), &header); err != nil {
		fmt.Println("DevInfoHandler header:", err)
	}
	//fmt.Printf("%+v\n", header)
	
	if err := json.Unmarshal([]byte(contentRaw), &content); err != nil {
		fmt.Println("DevInfoHandler content:", err)
	}
	
	// get module name
	for module, param := range content {
		fmt.Println("DevInfoHandler: module=", module)
		
		// parse to strcut
		switch module {
		case MODULE_PROCESS:
			s.agentDriver.HandleProcess(param)

		case MODULE_DATA_SYNC:
			info := ContentDataSync{}
			if err := json.Unmarshal([]byte(*param), &info); err != nil {
				fmt.Println("DevInfoHandler DataSync:", err)
			}
			fmt.Printf("DevInfoHandler: %+v\n", info)			
		default:
			//fmt.Printf("DevInfoHandler: unknown module: [%s]\n", module)
			//fmt.Printf("DevInfoHandler: payload: %s\n", string(msg.Payload()))
		}
	}
}

func (s *WiseAgentDriver) ParseCapability(contentRaw json.RawMessage) {
	var content map[string]*json.RawMessage

	if err := json.Unmarshal([]byte(contentRaw), &content); err != nil {
		fmt.Println("ParseCapability:", err)
	}

	for module, param := range content {
		fmt.Printf("ParseCapability: module=%s\n", module)
		switch module {
		case MODULE_PROCESS:
			s.HandleProcess(param)

		case MODULE_DATA_SYNC:
			capability := ContentDataSync{}
			if err := json.Unmarshal([]byte(*param), &capability); err != nil {
				fmt.Println("ParseCapability DataSync:", err)
			}
			fmt.Printf("ParseCapability:[%s] %+v\n", module, capability)

		default:
			//fmt.Println("Unknown capability module: ", module)
		}
	}
}

func (s *EdgeXDriver) AgentActionHandler(client MQTT.Client, msg MQTT.Message) {
	var contentRaw json.RawMessage

	header := AgentHeader { Content: &contentRaw, }	
	
	// parse header
	if err := json.Unmarshal([]byte(msg.Payload()), &header); err != nil {
		fmt.Println("AgentActionHandler header:", err)
		return
	}
	
	fmt.Printf("AgentActionHandler: Get command %d\n", header.CommCmd)
	//fmt.Printf("Message: %s\n", string(contentRaw))

	switch header.CommCmd {
	case WISE_GET_CAPABILITY_REP:
		s.agentDriver.ParseCapability(contentRaw);

	case WISE_GET_SENSOR_DATA_REQ:
		content := ContentGetSensor{}
		if err := json.Unmarshal([]byte(contentRaw), &content); err != nil {
			fmt.Println("AgentActionHandler get sensor:", err)
			return
		}
		fmt.Printf("%+v\n", content)
	default:
		fmt.Println("AgentActionHandler: Unknown CommCmd")
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 3X protocol callback
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
func (s *EdgeXDriver) SendAgentActionReq_3X(agentID string, data interface{}) (error) {
	payload, _ := json.Marshal(data)
	topic := fmt.Sprintf(DEF_CALLBACKREQ_TOPIC_3X, agentID)
	token := s.agentDriver.mqttClient.Publish(topic, 0, false, string(payload))
	token.Wait()

	return nil;
}

func (s *EdgeXDriver) AgentInfoHandler_3X(client MQTT.Client, msg MQTT.Message) {
	content := AgentInfoAck_3X{}
	if err := json.Unmarshal([]byte(msg.Payload()), &content); err != nil {
		fmt.Println("AgentInfoHandler_3X:", err)
		return
	}
	
	//PrintMQTT(msg, agentinfoack)
	fmt.Println("AgentInfoHandler_3X: Device ", content.SusiCommData.AgentID, " online, MAC=", content.SusiCommData.Mac)
	
	// send capabiity requests
	fmt.Println("AgentInfoHandler_3X send capability request")
	capRequest := &RequestCapability_3X{}
	capRequest.SusiCommData.RequestID = 1001
	capRequest.SusiCommData.CatalogID = 4
	capRequest.SusiCommData.HandlerName = "general"
	capRequest.SusiCommData.CommCmd = WISE_GET_CAPABILITY_REQ
	s.SendAgentActionReq_3X(content.SusiCommData.AgentID, capRequest)

	// send start reort
	fmt.Println("AgentInfoHandler_3X send start report request")
	reportRequest := &RequestStartReport_3X{}
	reportRequest.SusiCommData.CatalogID = 4
	reportRequest.SusiCommData.AutoUploadIntervalSec = 10
	reportRequest.SusiCommData.HandlerName = "general"
	reportRequest.SusiCommData.CommCmd = WISE_START_REPORT_REQ
	s.SendAgentActionReq_3X(content.SusiCommData.AgentID, reportRequest)
}

func (s *WiseAgentDriver) HandleCapabilityProcess_3X(capability CapabilityProcess_3X) {
	for _, e := range capability.SusiCommData.InfoSpec.ProcessMonitor.SystemMonitorInfo.E {
		switch e.N {
		case "CPU Usage":
			s.data.cpu = uint8(e.V)
		case "totalPhysMemKB":
			s.data.totalPhysMemKB = e.V
		case "availPhysMemKB":
			s.data.availPhysMemKB = e.V
		default:
		}
	}
	fmt.Printf("HandleProcess_3X: [%s] CPU=%d, totalPhysMemKB=%d, availPhysMemKB=%d\n",
			MODULE_PROCESS,
			s.data.cpu,
			s.data.totalPhysMemKB,
			s.data.availPhysMemKB)
}

func (s *EdgeXDriver) AgentActionHandler_3X(client MQTT.Client, msg MQTT.Message) {
	capability := CapabilityProcess_3X{}

	// parse header
	if err := json.Unmarshal([]byte(msg.Payload()), &capability); err != nil {
		fmt.Println("AgentActionHandler_3X capability:", err)
		return
	}
	if (capability.SusiCommData.CommCmd != WISE_GET_CAPABILITY_REP) {
		fmt.Println("AgentActionHandler_3X commCmmd: unknown command")
		return
	}
	fmt.Printf("AgentActionHandler_3X: Get command %d\n", capability.SusiCommData.CommCmd)
	
	s.agentDriver.HandleCapabilityProcess_3X(capability)
}

func (s *WiseAgentDriver) HandleReportProcess_3X(report ReportProcess_3X) {
	for _, e := range report.SusiCommData.Data.ProcessMonitor.SystemMonitorInfo.E {
		switch e.N {
		case "CPU Usage":
			s.data.cpu = uint8(e.V)
		case "totalPhysMemKB":
			s.data.totalPhysMemKB = e.V
		case "availPhysMemKB":
			s.data.availPhysMemKB = e.V
		default:
		}
	}
	fmt.Printf("HandleProcess_3X: [%s] CPU=%d, totalPhysMemKB=%d, availPhysMemKB=%d\n",
			MODULE_PROCESS,
			s.data.cpu,
			s.data.totalPhysMemKB,
			s.data.availPhysMemKB)
}

func (s *EdgeXDriver) DevInfoHandler_3X(client MQTT.Client, msg MQTT.Message) {
	report := ReportProcess_3X {}
	
	// parse header
	if err := json.Unmarshal([]byte(msg.Payload()), &report); err != nil {
		fmt.Println("DevInfoHandler_3X report:", err)
	}

	if (report.SusiCommData.CommCmd != WISE_REPORT_DATA_REP) {
		fmt.Println("DevInfoHandler_3X commCmmd: unknown command")
		return
	}
	
	fmt.Printf("DevInfoHandler_3X: Get command %d\n", report.SusiCommData.CommCmd)
	s.agentDriver.HandleReportProcess_3X(report)
}


func (s *EdgeXDriver) ConnectToBroker(broker string, clientId string) error {
	var topic string

	// init mqtt connection
	opts := MQTT.NewClientOptions()
	opts.AddBroker(broker)
	opts.SetClientID(clientId)
	opts.SetAutoReconnect(true)
	s.agentDriver = &WiseAgentDriver{}
	s.agentDriver.mqttClient = MQTT.NewClient(opts)
	
	// wait for connect
	if token := s.agentDriver.mqttClient.Connect(); token.Wait() && token.Error() != nil {
		s.lc.Error(fmt.Sprintf("MQTT connect fail: %v", token.Error()))
		return token.Error()
	}
	
	// 2.0
	if token := s.agentDriver.mqttClient.Subscribe(INFOACK_TOPIC, byte(0), s.AgentInfoHandler); token.Wait() && token.Error() != nil {
		s.lc.Error(fmt.Sprintf("MQTT subscribe %s fail: %v", INFOACK_TOPIC, token.Error()))
		return token.Error()
	}
	
	if token := s.agentDriver.mqttClient.Subscribe(DEF_AGENTREPORT_TOPIC, byte(0), s.DevInfoHandler); token.Wait() && token.Error() != nil {
		s.lc.Error(fmt.Sprintf("MQTT subscribe %s fail: %v", DEF_AGENTREPORT_TOPIC, token.Error()))
		return token.Error()
	}
	
	topic = fmt.Sprintf(DEF_AGENTACT_TOPIC, "RMM")
	if token := s.agentDriver.mqttClient.Subscribe(topic, byte(0), s.AgentActionHandler); token.Wait() && token.Error() != nil {
		s.lc.Error(fmt.Sprintf("MQTT subscribe %s fail: %v", topic, token.Error()))
		return token.Error()
	}
	
	// 3.3
	if token := s.agentDriver.mqttClient.Subscribe(INFOACK_TOPIC_3X, byte(0), s.AgentInfoHandler_3X); token.Wait() && token.Error() != nil {
		s.lc.Error(fmt.Sprintf("MQTT subscribe %s fail: %v", INFOACK_TOPIC, token.Error()))
		return token.Error()
	}

	if token := s.agentDriver.mqttClient.Subscribe(DEF_AGENTACT_TOPIC_3X, byte(0), s.AgentActionHandler_3X); token.Wait() && token.Error() != nil {
		s.lc.Error(fmt.Sprintf("MQTT subscribe %s fail: %v", DEF_AGENTACT_TOPIC_3X, token.Error()))
		return token.Error()
	}

	if token := s.agentDriver.mqttClient.Subscribe(DEF_AGENTREPORT_TOPIC_3X, byte(0), s.DevInfoHandler_3X); token.Wait() && token.Error() != nil {
		s.lc.Error(fmt.Sprintf("MQTT subscribe %s fail: %v", DEF_AGENTREPORT_TOPIC, token.Error()))
		return token.Error()
	}

	return nil
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// EdgeX callback implement
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

// DisconnectDevice handles protocol-specific cleanup when a device
// is removed.
func (s *EdgeXDriver) DisconnectDevice(address *models.Addressable) error {
	return nil
}

// Initialize performs protocol-specific initialization for the device
// service.
func (s *EdgeXDriver) Initialize(lc logger.LoggingClient, asyncCh chan<- *ds_models.AsyncValues) error {
	s.lc = lc
	s.asyncCh = asyncCh
	
	s.ConnectToBroker("tcp://127.0.0.1:1883", "wise-agent-sample")

	return nil
}

// HandleReadCommands triggers a protocol Read operation for the specified device.
func (s *EdgeXDriver) HandleReadCommands(addr *models.Addressable, reqs []ds_models.CommandRequest) (res []*ds_models.CommandValue, err error) {
	/*
	if len(reqs) > 1 {
		err = fmt.Errorf("EdgeXDriver.HandleReadCommands; too many command requests; only one supported") // this message will return to Rest API client
		return
	}
	*/

	s.lc.Debug(fmt.Sprintf("EdgeXDriver.HandleReadCommands: device: %s operation: %v attributes: %v", addr.Name, reqs[0].RO.Operation, reqs[0].DeviceObject.Attributes))
	res = make([]*ds_models.CommandValue, 3)
	now := time.Now().UnixNano() / int64(time.Millisecond)
	cpu, _ := ds_models.NewUint8Value(&reqs[0].RO, now, s.agentDriver.data.cpu)
	res[0] = cpu
	
	total, _ := ds_models.NewUint64Value(&reqs[1].RO, now, s.agentDriver.data.totalPhysMemKB)
	res[1] = total
	
	available, _ := ds_models.NewUint64Value(&reqs[2].RO, now, s.agentDriver.data.availPhysMemKB)
	res[2] = available

	return
}

// HandleWriteCommands passes a slice of CommandRequest struct each representing
// a ResourceOperation for a specific device resource (aka DeviceObject).
// Since the commands are actuation commands, params provide parameters for the individual
// command.
func (s *EdgeXDriver) HandleWriteCommands(addr *models.Addressable, reqs []ds_models.CommandRequest,
	params []*ds_models.CommandValue) error {

	if len(reqs) != 1 {
		err := fmt.Errorf("EdgeXDriver.HandleWriteCommands; too many command requests; only one supported")
		return err
	}
	if len(params) != 1 {
		err := fmt.Errorf("EdgeXDriver.HandleWriteCommands; the number of parameter is not correct; only one supported")
		return err
	}

	s.lc.Debug(fmt.Sprintf("EdgeXDriver.HandleWriteCommands: device: %s, operation: %v, parameters: %v", addr.Name, reqs[0].RO.Operation, params))
	var err error
	if s.switchButton, err = params[0].BoolValue(); err != nil {
		err := fmt.Errorf("EdgeXDriver.HandleWriteCommands; the data type of parameter should be Boolean, parameter: %s", params[0].String())
		return err
	}

	return nil
}

// Stop the protocol-specific DS code to shutdown gracefully, or
// if the force parameter is 'true', immediately. The driver is responsible
// for closing any in-use channels, including the channel used to send async
// readings (if supported).
func (s *EdgeXDriver) Stop(force bool) error {
	s.lc.Debug(fmt.Sprintf("EdgeXDriver.Stop called: force=%v", force))
	
	s.agentDriver.mqttClient.Disconnect(250)
	
	return nil
}
